from django.db.models import Q

from guardian.shortcuts import assign_perm, remove_perm

from huscy.projects.models import Membership, Project


def add_member(project, user, is_coordinator=False, has_write_permission=False):
    assign_perm('view_project', user, project)
    if is_coordinator or has_write_permission:
        assign_perm('change_project', user, project)
    return Membership.objects.create(
        project=project,
        user=user,
        is_coordinator=is_coordinator,
    )


def get_memberships(project=None):
    qs = Membership.objects.order_by('project__id', 'user__id')
    if project:
        qs = qs.filter(project=project)
    return qs


def remove_member(membership):
    remove_perm('view_project', membership.user, membership.project)
    remove_perm('change_project', membership.user, membership.project)
    membership.delete()


def update_membership(membership, is_coordinator, has_write_permission):
    if is_coordinator or has_write_permission:
        assign_perm('change_project', membership.user, membership.project)
    else:
        remove_perm('change_project', membership.user, membership.project)
    membership.is_coordinator = is_coordinator
    membership.save()
    return membership


def get_participating_projects(user):
    return (Project.objects
                   .filter(Q(principal_investigator=user) | Q(membership__user=user))
                   .distinct())



from rest_framework.routers import DefaultRouter

from huscy.projects import views


router = DefaultRouter()
router.register('dataacquisitionmethods', views.DataAcquisitionMethodViewSet)
router.register('experiments', views.ExperimentViewSet)
router.register('memberships', views.MembershipViewSet)
router.register('projects', views.ProjectViewSet, basename='project')
router.register('researchunits', views.ResearchUnitViewSet)
router.register('sessions', views.SessionViewSet)

urlpatterns = router.urls



from django.contrib import admin

from huscy.projects import models


admin.site.register(models.Experiment)
admin.site.register(models.Project)
admin.site.register(models.ResearchUnit)
admin.site.register(models.Session)



import pytest

from rest_framework.reverse import reverse
from rest_framework.status import HTTP_204_NO_CONTENT, HTTP_403_FORBIDDEN

from django.contrib.auth.models import Permission

pytestmark = pytest.mark.django_db


def test_admin_user_can_delete_projects(admin_client, session):
    response = delete_session(admin_client, session)

    assert response.status_code == HTTP_204_NO_CONTENT


def test_user_with_permission_can_delete_projects(client, user, session):
    delete_permission = Permission.objects.get(codename='delete_session',
                                               content_type__app_label='projects')
    user.user_permissions.add(delete_permission)

    response = delete_session(client, session)

    assert response.status_code == HTTP_204_NO_CONTENT


def test_user_without_permission_can_delete_projects(client, session):
    response = delete_session(client, session)

    assert response.status_code == HTTP_204_NO_CONTENT


def test_anonymous_user_cannot_delete_projects(anonymous_client, session):
    response = delete_session(anonymous_client, session)

    assert response.status_code == HTTP_403_FORBIDDEN


def delete_session(client, session):
    return client.delete(reverse('session-detail', kwargs=dict(pk=session.pk)))



from model_bakery import baker

from huscy.projects.models import Membership, Project
from huscy.projects.services import create_project

public = Project.VISIBILITY.get_value('public')


def test_creator_is_principal_investigator(user, research_unit):
    project = create_project('title', research_unit, user, user, public)

    assert Project.objects.exists()
    assert not Membership.objects.exists()
    assert project.description == ''


def test_creator_is_not_principal_investigator(django_user_model, user, research_unit):
    principal_investigator = baker.make(django_user_model)
    project = create_project('title', research_unit, principal_investigator, user, public)

    assert Project.objects.exists()
    assert Membership.objects.filter(user=user, is_coordinator=True).exists()
    assert project.description == ''


def test_with_optional_description(user, research_unit):
    project = create_project('title', research_unit, user, user, public, description='description')

    assert Project.objects.exists()
    assert not project.description == ''


def test_with_optional_local_id(user, research_unit):
    project = create_project('title', research_unit, user, user, public, local_id=166)
    assert Project.objects.exists()
    assert project.local_id == 166



import pytest

from rest_framework.reverse import reverse
from rest_framework.status import HTTP_201_CREATED, HTTP_403_FORBIDDEN

from huscy.projects import services

pytestmark = pytest.mark.django_db


def test_admin_user_can_create_data_acquisition_methods(admin_client, session):
    response = create_data_acquisition_method(admin_client, session)

    assert response.status_code == HTTP_201_CREATED


def test_user_without_permission_can_create_data_acquisition_methods(client, session):
    response = create_data_acquisition_method(client, session)

    assert response.status_code == HTTP_201_CREATED


def test_anonymous_user_cannot_create_data_acquisition_methods(anonymous_client, session):
    response = create_data_acquisition_method(anonymous_client, session)

    assert response.status_code == HTTP_403_FORBIDDEN


@pytest.mark.skip
def test_service_function_was_called(mocker, client, session):
    spy = mocker.spy(services, 'add_data_acquisition_method')

    create_data_acquisition_method(client, session)

    spy.assert_called_once_with(session, 'eeg', 'C206', 1)


def create_data_acquisition_method(client, session):
    return client.post(
        reverse('dataacquisitionmethod-list'),
        data=dict(session=session.id, order=1, type='eeg', location='C206')
    )



import logging

from rest_framework import serializers

from .membership import MembershipSerializer
from huscy.projects.helpers import get_client_ip
from huscy.projects.models import Project
from huscy.projects.services import create_project, get_memberships

logger = logging.getLogger('projects')


class ProjectSerializer(serializers.ModelSerializer):
    creator = serializers.HiddenField(default=serializers.CurrentUserDefault())
    members = serializers.SerializerMethodField()
    principal_investigator_name = serializers.SerializerMethodField()
    research_unit_name = serializers.SerializerMethodField()

    # default=None is used as required=False does not work together with the unique_together
    # constraint, however this defaut=None is a workaround, see:
    # https://github.com/encode/django-rest-framework/issues/4456
    local_id = serializers.IntegerField(min_value=1, default=None)

    class Meta:
        model = Project
        fields = (
            'creator',
            'description',
            'id',
            'local_id',
            'local_id_name',
            'members',
            'principal_investigator',
            'principal_investigator_name',
            'research_unit',
            'research_unit_name',
            'title',
            'visibility',
        )

    def get_members(self, project):
        members = get_memberships(project)
        return MembershipSerializer(members, many=True).data

    def get_principal_investigator_name(self, project):
        return project.principal_investigator.get_full_name()

    def get_research_unit_name(self, project):
        return project.research_unit.name

    def create(self, validated_data):
        request = self.context.get('request')
        logger.info('User %s from IP %s requested creation of new project',
                    request.user.username, get_client_ip(request))
        return create_project(**validated_data)



# Generated by Django 3.0.8 on 2020-07-15 09:05

import datetime
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import huscy.projects.models.reserach_unit


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='DataAcquisitionMethod',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('order', models.PositiveSmallIntegerField()),
                ('type', models.CharField(choices=[('behav', 'Behavioral'), ('bio', 'Biological samples'), ('eeg', 'Electroencephalography'), ('meg', 'Magnetoencephalography'), ('micro', 'Microscopy data'), ('mri', 'Magnetic resonance imaging'), ('nirs', 'Near-Infrared Spectroscopy'), ('pause', 'Pause'), ('pet', 'Positron-emission tomography'), ('phys', 'Physiological measures'), ('quest', 'Questionnaire'), ('screen', 'Screening')], max_length=16)),
                ('location', models.CharField(max_length=126)),
            ],
        ),
        migrations.CreateModel(
            name='Experiment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=64)),
                ('description', models.TextField(blank=True, default='')),
                ('order', models.PositiveSmallIntegerField()),
            ],
        ),
        migrations.CreateModel(
            name='Stimulus',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('type', models.PositiveSmallIntegerField(choices=[(0, 'auditive'), (1, 'gustatory'), (2, 'haptic'), (3, 'olfactory'), (4, 'visual')])),
            ],
        ),
        migrations.CreateModel(
            name='Session',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=64)),
                ('order', models.PositiveSmallIntegerField()),
                ('setup_time', models.DurationField(default=datetime.timedelta(0))),
                ('duration', models.DurationField()),
                ('teardown_time', models.DurationField(default=datetime.timedelta(0))),
                ('max_number_of_participants', models.PositiveIntegerField(default=1)),
                ('experiment', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='sessions', to='projects.Experiment')),
                ('operator', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='Screening',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('method', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='screenings', to='projects.DataAcquisitionMethod')),
            ],
        ),
        migrations.CreateModel(
            name='ResearchUnit',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=128, unique=True)),
                ('code', models.CharField(blank=True, default=huscy.projects.models.reserach_unit.get_default_value_for_code, max_length=32, unique=True)),
                ('principal_investigator', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name='Questionnaire',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('method', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='questionnaires', to='projects.DataAcquisitionMethod')),
            ],
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('local_id', models.PositiveIntegerField()),
                ('title', models.CharField(max_length=126)),
                ('description', models.TextField(blank=True, default='')),
                ('visibility', models.PositiveSmallIntegerField(choices=[(0, 'public'), (1, 'private')], default=0)),
                ('principal_investigator', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to=settings.AUTH_USER_MODEL)),
                ('research_unit', models.ForeignKey(on_delete=django.db.models.deletion.DO_NOTHING, to='projects.ResearchUnit')),
            ],
            options={
                'permissions': (('can_view_private_projects', 'Can view private projects'),),
                'unique_together': {('local_id', 'research_unit')},
            },
        ),
        migrations.CreateModel(
            name='Physiological',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('method', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='physiologicals', to='projects.DataAcquisitionMethod')),
            ],
        ),
        migrations.CreateModel(
            name='PET',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('method', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='pet', to='projects.DataAcquisitionMethod')),
            ],
        ),
        migrations.CreateModel(
            name='Pause',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('method', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='pause', to='projects.DataAcquisitionMethod')),
            ],
        ),
        migrations.CreateModel(
            name='NIRS',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('method', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='nirs', to='projects.DataAcquisitionMethod')),
            ],
        ),
        migrations.CreateModel(
            name='MRI',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('method', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='mri', to='projects.DataAcquisitionMethod')),
            ],
        ),
        migrations.CreateModel(
            name='Microscopy',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('method', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='microscopys', to='projects.DataAcquisitionMethod')),
            ],
        ),
        migrations.CreateModel(
            name='MEG',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('method', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='meg', to='projects.DataAcquisitionMethod')),
            ],
        ),
        migrations.AddField(
            model_name='experiment',
            name='project',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='experiments', to='projects.Project'),
        ),
        migrations.CreateModel(
            name='EEG',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('method', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='eeg', to='projects.DataAcquisitionMethod')),
            ],
        ),
        migrations.AddField(
            model_name='dataacquisitionmethod',
            name='session',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='projects.Session'),
        ),
        migrations.AddField(
            model_name='dataacquisitionmethod',
            name='stimulus',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='projects.Stimulus'),
        ),
        migrations.CreateModel(
            name='Biological',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('method', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='biologicals', to='projects.DataAcquisitionMethod')),
            ],
        ),
        migrations.CreateModel(
            name='Behavioral',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('method', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='behaviorals', to='projects.DataAcquisitionMethod')),
            ],
        ),
        migrations.CreateModel(
            name='Membership',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_coordinator', models.BooleanField(default=False)),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='projects.Project')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'unique_together': {('user', 'project')},
            },
        ),
    ]



from .data_acquisition_method import DataAcquisitionMethodSerializer
from .experiment import ExperimentSerializer
from .membership import MembershipSerializer
from .project import ProjectSerializer
from .research_unit import ResearchUnitSerializer
from .session import SessionSerializer


__all__ = (
    'DataAcquisitionMethodSerializer',
    'ExperimentSerializer',
    'MembershipSerializer',
    'ProjectSerializer',
    'ResearchUnitSerializer',
    'SessionSerializer',
)



from itertools import cycle

import pytest
from model_bakery import baker

from rest_framework.reverse import reverse
from rest_framework.status import HTTP_200_OK, HTTP_403_FORBIDDEN

from huscy.projects import services

pytestmark = pytest.mark.django_db


def test_get_research_units_service_function():
    names = ['Project1', 'This is a project', 'A new project']
    baker.make('projects.ResearchUnit', name=cycle(names), _quantity=len(names))

    research_units = services.get_research_units()

    assert sorted(names) == list(research_units.values_list('name', flat=True))


def test_admin_user_can_list_research_units(admin_client):
    response = list_research_unit(admin_client)

    assert response.status_code == HTTP_200_OK


def test_user_without_permission_can_list_research_units(client):
    response = list_research_unit(client)

    assert response.status_code == HTTP_200_OK


def test_anonymous_user_cannot_list_research_units(anonymous_client):
    response = list_research_unit(anonymous_client)

    assert response.status_code == HTTP_403_FORBIDDEN


def list_research_unit(client):
    return client.get(reverse('researchunit-list'))



from uuid import uuid4

from django.conf import settings
from django.db import models


def get_default_value_for_code():
    return uuid4().hex


class ResearchUnit(models.Model):
    name = models.CharField(max_length=128, unique=True)
    code = models.CharField(max_length=32, unique=True, blank=True,
                            default=get_default_value_for_code)
    principal_investigator = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.PROTECT)

    def __str__(self):
        return f'{self.name} ({self.principal_investigator.get_full_name()})'



from django.db import models

from .project import Project


class Experiment(models.Model):
    project = models.ForeignKey(Project, on_delete=models.CASCADE, related_name='experiments')

    title = models.CharField(max_length=64)
    description = models.TextField(blank=True, default='')

    order = models.PositiveSmallIntegerField()



from rest_framework.reverse import reverse
from rest_framework.status import HTTP_200_OK, HTTP_403_FORBIDDEN


def test_admin_user_can_update_memberships(admin_client, membership):
    assert not membership.user.has_perm('change_project', membership.project)

    response = update_membership(admin_client, membership)

    assert response.status_code == HTTP_200_OK
    assert membership.user.has_perm('change_project', membership.project)


def test_user_without_permission_can_update_memberships(client, membership):
    assert not membership.user.has_perm('change_project', membership.project)

    response = update_membership(client, membership)

    assert response.status_code == HTTP_200_OK
    assert membership.user.has_perm('change_project', membership.project)


def test_anonymous_user_cannot_update_memberships(anonymous_client, membership):
    assert not membership.user.has_perm('change_project', membership.project)

    response = update_membership(anonymous_client, membership)

    assert response.status_code == HTTP_403_FORBIDDEN
    assert not membership.user.has_perm('change_project', membership.project)


def update_membership(client, membership):
    return client.put(
        reverse('membership-detail', kwargs=dict(pk=membership.id)),
        data=dict(
            project=membership.project.id,
            user=membership.user.id,
            is_coordinator=False,
            has_write_permission=True,
        ),
    )



from huscy.projects.filters import DataAcquisitionMethodFilter
from huscy.projects.models import DataAcquisitionMethod


def add_data_acquisition_method(session, type, location, order=None):
    if order is None:
        order = DataAcquisitionMethod.objects.filter(session=session).count() + 1

    return DataAcquisitionMethod.objects.create(
        session=session,
        type=type,
        location=location,
        order=order
    )


def get_data_acquisition_methods(project=None):
    qs = DataAcquisitionMethod.objects.order_by('pk')
    filters = dict(project=project and project.pk)
    return DataAcquisitionMethodFilter(filters, queryset=qs).qs



from rest_framework.reverse import reverse
from rest_framework.status import HTTP_405_METHOD_NOT_ALLOWED


def test_list_memberships_not_provided(admin_client):
    response = list_memberships(admin_client)

    assert_method_not_allowed(response)


def test_retrieve_memberships_not_provided(client, membership):
    response = retrieve_membership(client, membership)

    assert_method_not_allowed(response)


def list_memberships(client):
    return client.get(reverse('membership-list'))


def retrieve_membership(client, membership):
    return client.get(reverse('membership-detail', kwargs=dict(pk=membership.pk)))


def assert_method_not_allowed(response):
    assert response.status_code == HTTP_405_METHOD_NOT_ALLOWED



from itertools import cycle

import pytest
from model_bakery import baker

from huscy.projects.services import get_experiments

pytestmark = pytest.mark.django_db


@pytest.fixture
def experiments():
    projects = baker.make('projects.Project', _quantity=3)
    return baker.make('projects.Experiment', project=cycle(projects), _quantity=6)


def test_get_experiments(experiments):
    result = get_experiments()

    assert list(result) == experiments


def test_get_experiments_filtered_by_project(experiments):
    result = get_experiments(experiments[0].project)

    assert len(result) == 2
    assert list(result) == [experiments[0], experiments[3]]



from rest_framework.reverse import reverse
from rest_framework.status import HTTP_201_CREATED, HTTP_403_FORBIDDEN

from django.contrib.auth.models import Permission


def test_admin_user_can_create_research_units(admin_client, user):
    response = create_research_unit(admin_client, user)

    assert response.status_code == HTTP_201_CREATED


def test_user_with_permission_can_create_research_units(client, user):
    create_permission = Permission.objects.get(codename='add_researchunit')
    user.user_permissions.add(create_permission)

    response = create_research_unit(client, user)

    assert response.status_code == HTTP_201_CREATED, response.json()


def test_user_without_permission_cannot_create_research_units(client, user):
    response = create_research_unit(client, user)

    assert response.status_code == HTTP_403_FORBIDDEN


def test_anonymous_user_cannot_create_research_units(anonymous_client, user):
    response = create_research_unit(anonymous_client, user)

    assert response.status_code == HTTP_403_FORBIDDEN


def create_research_unit(client, user):
    return client.post(
        reverse('researchunit-list'),
        data=dict(name='research_unit_1', code='RU1', principal_investigator=user.pk)
    )



from django_filters import rest_framework as filters

from huscy.projects import models


class DataAcquisitionMethodFilter(filters.FilterSet):
    project = filters.NumberFilter(field_name='session__experiment__project', lookup_expr='exact')

    class Meta:
        model = models.DataAcquisitionMethod
        fields = (
            'project',
        )


class ExperimentFilter(filters.FilterSet):
    class Meta:
        model = models.Experiment
        fields = (
            'project',
        )



import pytest

from rest_framework.reverse import reverse
from rest_framework.status import HTTP_204_NO_CONTENT, HTTP_403_FORBIDDEN

pytestmark = pytest.mark.django_db


def test_admin_user_can_delete_experiments(admin_client, experiment):
    response = delete_experiment(admin_client, experiment)

    assert response.status_code == HTTP_204_NO_CONTENT


def test_user_without_permission_can_delete_experiments(client, experiment):
    response = delete_experiment(client, experiment)

    assert response.status_code == HTTP_204_NO_CONTENT


def test_anonymous_user_cannot_delete_experiments(anonymous_client, experiment):
    response = delete_experiment(anonymous_client, experiment)

    assert response.status_code == HTTP_403_FORBIDDEN


def delete_experiment(client, experiment):
    return client.delete(reverse('experiment-detail', kwargs=dict(pk=experiment.pk)))



from model_bakery import baker

from huscy.projects.serializer import MembershipSerializer, ProjectSerializer
from huscy.projects.services import add_member


def test_project_serializer(django_user_model, project):
    membership1 = add_member(project, baker.make(django_user_model), is_coordinator=True)
    membership2 = add_member(project, baker.make(django_user_model))

    expected = dict(
        description=project.description,
        id=project.id,
        local_id=project.local_id,
        local_id_name=project.local_id_name,
        members=[
            MembershipSerializer(membership1).data,
            MembershipSerializer(membership2).data,
        ],
        principal_investigator=project.principal_investigator.pk,
        principal_investigator_name=project.principal_investigator.get_full_name(),
        research_unit=project.research_unit.pk,
        research_unit_name=project.research_unit.name,
        title=project.title,
        visibility=project.visibility,
    )

    assert expected == ProjectSerializer(instance=project).data



from .data_acquisition_method import DataAcquisitionMethod
from .experiment import Experiment
from .membership import Membership
from .project import Project
from .reserach_unit import ResearchUnit
from .session import Session
from .stimulus import Stimulus

__all__ = (
    'DataAcquisitionMethod',
    'Experiment',
    'Membership',
    'Project',
    'ResearchUnit',
    'Session',
    'Stimulus',
)



from datetime import timedelta

import pytest

from rest_framework.reverse import reverse
from rest_framework.status import HTTP_201_CREATED, HTTP_403_FORBIDDEN

from django.contrib.auth.models import Permission

pytestmark = pytest.mark.django_db


def test_admin_user_can_create_sessions(admin_client, user, experiment):
    response = create_session(admin_client, experiment, user)

    assert response.status_code == HTTP_201_CREATED


def test_user_with_permission_can_create_sessions(client, user, experiment):
    create_permission = Permission.objects.get(codename='add_session',
                                               content_type__app_label='projects')
    user.user_permissions.add(create_permission)

    response = create_session(client, experiment, user)

    assert response.status_code == HTTP_201_CREATED


def test_user_without_permission_can_create_sessions(client, user, experiment):
    response = create_session(client, experiment, user)

    assert response.status_code == HTTP_201_CREATED


def test_anonymous_user_cannot_create_sessions(anonymous_client, user, experiment):
    response = create_session(anonymous_client, experiment, user)

    assert response.status_code == HTTP_403_FORBIDDEN


def create_session(client, experiment, operator):
    return client.post(
        reverse('session-list'),
        data=dict(
            experiment=experiment.pk,
            operator=operator.pk,
            duration=timedelta(hours=1),
        )
    )



from rest_framework import serializers

from .session import SessionSerializer
from huscy.projects.models import Experiment
from huscy.projects.services import create_experiment


class ExperimentSerializer(serializers.ModelSerializer):
    order = serializers.IntegerField(required=False)
    title = serializers.CharField(required=False)

    class Meta:
        model = Experiment
        fields = (
            'description',
            'id',
            'order',
            'project',
            'title',
        )

    def create(self, validated_data):
        return create_experiment(**validated_data)

    def to_representation(self, experiment):
        response = super().to_representation(experiment)
        response['sessions'] = SessionSerializer(experiment.sessions.all(), many=True).data
        return response



import pytest

from rest_framework.reverse import reverse
from rest_framework.status import HTTP_200_OK, HTTP_403_FORBIDDEN


def test_admin_user_can_list_sessions(admin_client):
    response = list_sessions(admin_client)

    assert_success(response)


def test_user_without_permission_can_list_sessions(client):
    response = list_sessions(client)

    assert_success(response)


@pytest.mark.django_db
def test_anonymous_user_cannot_list_sessions(anonymous_client):
    response = list_sessions(anonymous_client)

    assert_permission_denied(response)


def retrieve_session(client, session):
    return client.get(reverse('session-detail', kwargs=dict(pk=session.id)))


def list_sessions(client):
    return client.get(reverse('session-list'))


def assert_success(response):
    assert response.status_code == HTTP_200_OK


def assert_permission_denied(response):
    assert response.status_code == HTTP_403_FORBIDDEN



import pytest

from rest_framework.reverse import reverse
from rest_framework.status import HTTP_200_OK, HTTP_403_FORBIDDEN

from huscy.projects import services

pytestmark = pytest.mark.django_db


def test_admin_user_can_get_experiments(admin_client, experiment):
    response = list_experiments(admin_client, experiment.project)

    assert response.status_code == HTTP_200_OK


def test_user_without_permission_can_get_experiments(client, experiment):
    response = list_experiments(client, experiment.project)

    assert response.status_code == HTTP_200_OK


def test_anonymous_user_cannot_get_experiments(anonymous_client, experiment):
    response = list_experiments(anonymous_client, experiment.project)

    assert response.status_code == HTTP_403_FORBIDDEN


def test_service_function_was_called(mocker, client, project):
    spy = mocker.spy(services, 'get_experiments')

    list_experiments(client, project)

    spy.assert_called_once_with(project)


def list_experiments(client, project):
    return client.get(reverse('project-experiments', kwargs=dict(pk=project.pk)))



from rest_framework.reverse import reverse
from rest_framework.status import HTTP_204_NO_CONTENT, HTTP_403_FORBIDDEN

from django.contrib.auth.models import Permission


def test_admin_user_can_delete_research_units(admin_client, research_unit):
    response = delete_research_unit(admin_client, research_unit)

    assert response.status_code == HTTP_204_NO_CONTENT


def test_user_with_permission_can_delete_research_units(client, user, research_unit):
    create_permission = Permission.objects.get(codename='delete_researchunit')
    user.user_permissions.add(create_permission)

    response = delete_research_unit(client, research_unit)

    assert response.status_code == HTTP_204_NO_CONTENT


def test_user_without_permission_cannot_delete_research_units(client, research_unit):
    response = delete_research_unit(client, research_unit)

    assert response.status_code == HTTP_403_FORBIDDEN


def test_anonymous_user_cannot_delete_research_units(anonymous_client, research_unit):
    response = delete_research_unit(anonymous_client, research_unit)

    assert response.status_code == HTTP_403_FORBIDDEN


def delete_research_unit(client, research_unit):
    return client.delete(reverse('researchunit-detail', kwargs=dict(pk=research_unit.id)))



from rest_framework import serializers

from huscy.projects.models import Session
from huscy.projects.serializer import DataAcquisitionMethodSerializer
from huscy.projects.services import create_session


class SessionSerializer(serializers.ModelSerializer):
    order = serializers.IntegerField(required=False)
    title = serializers.CharField(required=False)

    class Meta:
        model = Session
        fields = (
            'duration',
            'experiment',
            'id',
            'operator',
            'order',
            'title',
        )

    def create(self, validated_data):
        return create_session(**validated_data)

    def to_representation(self, session):
        response = super().to_representation(session)
        response['data_acquisition_methods'] = \
            DataAcquisitionMethodSerializer(session.dataacquisitionmethod_set.all(), many=True).data
        return response



import pytest
from model_bakery import baker

from huscy.projects.services import create_experiment

pytestmark = pytest.mark.django_db


def test_create_experiment(project):
    experiment = create_experiment(project)

    assert experiment.project == project
    assert experiment.title == 'Experiment 1'
    assert experiment.description == ''
    assert experiment.order == 1


def test_create_experiment_to_existing_ones(project):
    baker.make('projects.Experiment', project=project, _quantity=2)
    experiment = create_experiment(project)

    assert experiment.title == 'Experiment 3'
    assert experiment.order == 3


def test_create_experiment_with_title(project):
    experiment = create_experiment(project, 'the experiment')

    assert experiment.title == 'the experiment'


def test_create_experiment_with_order(project):
    experiment = create_experiment(project, order=5)

    assert experiment.order == 5



from model_bakery import baker

from huscy.projects.serializer import ExperimentSerializer, SessionSerializer


def test_sessions_are_serialized_in_experiment_serializer(client, experiment):
    sessions = baker.make('projects.Session', experiment=experiment, _quantity=2)

    serialized_experiment = ExperimentSerializer(experiment).data

    assert serialized_experiment['sessions'] == SessionSerializer(sessions, many=True).data



import logging

from rest_framework import mixins, viewsets
from rest_framework.decorators import action
from rest_framework.permissions import DjangoModelPermissions, IsAuthenticated
from rest_framework.response import Response

from huscy.projects import helpers, permissions, serializer, services

logger = logging.getLogger('projects')


class DataAcquisitionMethodViewSet(mixins.CreateModelMixin, mixins.DestroyModelMixin,
                                   mixins.UpdateModelMixin, viewsets.GenericViewSet):
    queryset = services.get_data_acquisition_methods()
    serializer_class = serializer.DataAcquisitionMethodSerializer
    permission_classes = (IsAuthenticated, )


class ExperimentViewSet(mixins.CreateModelMixin, mixins.DestroyModelMixin, mixins.UpdateModelMixin,
                        viewsets.GenericViewSet):
    queryset = services.get_experiments()
    serializer_class = serializer.ExperimentSerializer
    permission_classes = (IsAuthenticated, )


class MembershipViewSet(mixins.CreateModelMixin, mixins.DestroyModelMixin, mixins.UpdateModelMixin,
                        viewsets.GenericViewSet):
    queryset = services.get_memberships()
    serializer_class = serializer.MembershipSerializer
    permission_classes = (IsAuthenticated, )

    def perform_destroy(self, membership):
        services.remove_member(membership)


class ProjectViewSet(mixins.CreateModelMixin, mixins.DestroyModelMixin, mixins.ListModelMixin,
                     mixins.UpdateModelMixin, viewsets.GenericViewSet):
    serializer_class = serializer.ProjectSerializer
    permission_classes = (IsAuthenticated, DjangoModelPermissions | permissions.AllowAnyToCreate)

    def get_queryset(self):
        return services.get_projects(self.request.user)

    def destroy(self, request, *args, **kwargs):
        pk = kwargs.get('pk')
        logger.info('User %s from ip %s requested deletion of project with id:%d',
                    request.user.username, helpers.get_client_ip(request), pk)
        return super().destroy(request, args, kwargs)

    def perform_destroy(self, project):
        services.delete_project(project)

    @action(detail=True, methods=['get'])
    def dataacquisitionmethods(self, request, pk=None):
        methods = services.get_data_acquisition_methods(self.get_object())
        return Response(data=serializer.DataAcquisitionMethodSerializer(methods, many=True).data)

    @action(detail=True, methods=['get'])
    def experiments(self, request, pk=None):
        experiments = services.get_experiments(self.get_object())
        return Response(data=serializer.ExperimentSerializer(experiments, many=True).data)


class ResearchUnitViewSet(viewsets.ModelViewSet):
    permission_classes = (DjangoModelPermissions, )
    queryset = services.get_research_units()
    serializer_class = serializer.ResearchUnitSerializer


class SessionViewSet(viewsets.ModelViewSet):
    queryset = services.get_sessions()
    serializer_class = serializer.SessionSerializer
    permission_classes = (IsAuthenticated, )



from enum import Enum

from django.db import models


class Stimulus(models.Model):
    class TYPE(Enum):
        auditive = (0, 'auditive')
        gustatory = (1, 'gustatory')
        haptic = (2, 'haptic')
        olfactory = (3, 'olfactory')
        visual = (4, 'visual')

        @classmethod
        def get_value(cls, member):
            return cls[member].value[0]

    type = models.PositiveSmallIntegerField(choices=[x.value for x in TYPE])



from datetime import timedelta

import pytest

from rest_framework.reverse import reverse
from rest_framework.status import HTTP_200_OK, HTTP_403_FORBIDDEN

from django.contrib.auth.models import Permission


ONE_HOUR = timedelta(hours=1)
TWO_HOURS = timedelta(hours=2)


def test_admin_user_can_update_sessions(admin_client, session):
    response = update_session(admin_client, session)

    assert_success(response)
    assert_session_is_updated(session)


def test_admin_user_can_partial_update_sessions(admin_client, session):
    response = partial_update_session(admin_client, session)

    assert_success(response)
    assert_session_is_updated(session)


def test_user_with_permission_can_update_sessions(client, user, session):
    update_permission = Permission.objects.get(codename='change_session',
                                               content_type__app_label='projects')
    user.user_permissions.add(update_permission)

    response = update_session(client, session)

    assert_success(response)
    assert_session_is_updated(session)


def test_user_with_permission_can_partial_update_sessions(client, user, session):
    update_permission = Permission.objects.get(codename='change_session',
                                               content_type__app_label='projects')
    user.user_permissions.add(update_permission)

    response = partial_update_session(client, session)

    assert_success(response)
    assert_session_is_updated(session)


def test_user_without_permission_can_update_sessions(client, session):
    response = update_session(client, session)

    assert_success(response)
    assert_session_is_updated(session)


def test_user_without_permission_can_partial_update_sessions(client, session):
    response = partial_update_session(client, session)

    assert_success(response)
    assert_session_is_updated(session)


@pytest.mark.django_db
def test_anonymous_user_cannot_update_sessions(anonymous_client, session):
    response = update_session(anonymous_client, session)

    assert_permission_denied(response)
    assert_session_did_not_change(session)


@pytest.mark.django_db
def test_anonymous_user_cannot_partial_update_sessions(anonymous_client, session):
    response = partial_update_session(anonymous_client, session)

    assert_permission_denied(response)
    assert_session_did_not_change(session)


def update_session(client, session):
    return client.put(
        reverse('session-detail', kwargs=dict(pk=session.id)),
        data=dict(
            experiment=session.experiment.pk,
            order=session.order,
            title='new title',
            setup_time=session.setup_time,
            duration=TWO_HOURS,
            teardown_time=session.teardown_time,
            operator=session.operator.pk,
            max_number_of_participants=session.max_number_of_participants,
        )
    )


def partial_update_session(client, session):
    return client.patch(
        reverse('session-detail', kwargs=dict(pk=session.id)),
        data=dict(duration=TWO_HOURS, title='new title')
    )


def assert_success(response):
    assert response.status_code == HTTP_200_OK


def assert_session_is_updated(session):
    session.refresh_from_db()
    assert TWO_HOURS == session.duration
    assert 'new title' == session.title


def assert_permission_denied(response):
    assert response.status_code == HTTP_403_FORBIDDEN


def assert_session_did_not_change(session):
    session.refresh_from_db()
    assert ONE_HOUR == session.duration



import pytest
from model_bakery import baker

from rest_framework.reverse import reverse
from rest_framework.status import HTTP_201_CREATED, HTTP_403_FORBIDDEN

pytestmark = pytest.mark.django_db


@pytest.fixture
def member(django_user_model):
    return baker.make(django_user_model)


def test_admin_user_can_create_memberships(admin_client, project, member):
    response = create_membership(admin_client, project, member)

    assert response.status_code == HTTP_201_CREATED


def test_user_without_permission_can_create_memberships(client, project, member):
    response = create_membership(client, project, member)

    assert response.status_code == HTTP_201_CREATED


def test_anonymous_user_cannot_create_memberships(anonymous_client, project, member):
    response = create_membership(anonymous_client, project, member)

    assert response.status_code == HTTP_403_FORBIDDEN


def create_membership(client, project, member, is_coordinator=False):
    return client.post(
        reverse('membership-list'),
        data=dict(project=project.id, user=member.id, is_coordinator=is_coordinator)
    )



import pytest

from django.db import IntegrityError

from huscy.projects.models import Membership
from huscy.projects.services import add_member


@pytest.mark.django_db
def test_add_normal_member(project, user):
    assert_membership_does_not_exist()

    add_member(project, user)

    assert_membership_exists(project, user)
    assert_user_has_view_permission_for_project(project, user)
    assert_user_does_not_have_change_permission_for_project(project, user)


@pytest.mark.django_db
def test_add_member_as_coordinator(project, user):
    assert_membership_does_not_exist()

    add_member(project, user, is_coordinator=True)

    assert_membership_exists(project, user)
    assert_user_is_coordinator(project, user)
    assert_user_has_view_permission_for_project(project, user)
    assert_user_has_change_permission_for_project(project, user)


@pytest.mark.django_db
def test_add_member_with_write_permission(project, user):
    assert_membership_does_not_exist()

    add_member(project, user, has_write_permission=True)

    assert_membership_exists(project, user)
    assert_user_is_not_coordinator(project, user)
    assert_user_has_view_permission_for_project(project, user)
    assert_user_has_change_permission_for_project(project, user)


@pytest.mark.django_db
def test_has_write_permission_is_ignored_if_is_coordinator_is_set_to_true(project, user):
    assert_membership_does_not_exist()

    add_member(project, user, is_coordinator=True, has_write_permission=False)

    assert_membership_exists(project, user)
    assert_user_is_coordinator(project, user)
    assert_user_has_view_permission_for_project(project, user)
    assert_user_has_change_permission_for_project(project, user)


@pytest.mark.django_db
def test_add_member_twice_to_project_fails(project, user):
    assert_membership_does_not_exist()

    add_member(project, user)
    with pytest.raises(IntegrityError):
        add_member(project, user)


def assert_membership_does_not_exist():
    assert not Membership.objects.exists()


def assert_membership_exists(project, user):
    membership = Membership.objects.get()
    assert membership.project == project
    assert membership.user == user


def assert_user_is_coordinator(project, user):
    membership = Membership.objects.filter(is_coordinator=True).get()
    assert membership.project == project
    assert membership.user == user


def assert_user_is_not_coordinator(project, user):
    membership = Membership.objects.filter(is_coordinator=False).get()
    assert membership.project == project
    assert membership.user == user


def assert_user_has_view_permission_for_project(project, user):
    assert user.has_perm('view_project', project)


def assert_user_has_change_permission_for_project(project, user):
    assert user.has_perm('change_project', project)


def assert_user_does_not_have_change_permission_for_project(project, user):
    assert not user.has_perm('change_project', project)



import pytest

from rest_framework.reverse import reverse
from rest_framework.status import HTTP_201_CREATED, HTTP_403_FORBIDDEN

from huscy.projects import services

pytestmark = pytest.mark.django_db


def test_admin_user_can_create_experiments(admin_client, project):
    response = create_experiment(admin_client, project)

    assert response.status_code == HTTP_201_CREATED


def test_user_without_permission_can_create_experiments(client, project):
    response = create_experiment(client, project)

    assert response.status_code == HTTP_201_CREATED


def test_anonymous_user_cannot_create_experiments(anonymous_client, project):
    response = create_experiment(anonymous_client, project)

    assert response.status_code == HTTP_403_FORBIDDEN


@pytest.mark.skip
def test_service_function_was_called(mocker, client, project):
    spy = mocker.spy(services, 'create_experiment')

    create_experiment(client, project)

    spy.assert_called_once_with(project)


def create_experiment(client, project):
    return client.post(reverse('experiment-list'), data=dict(project=project.id))



from huscy.projects.filters import ExperimentFilter
from huscy.projects.models import Experiment


def create_experiment(project, title='', description='', order=0):
    project_experiment_count = project.experiments.count()

    return Experiment.objects.create(
        description=description,
        order=order or project_experiment_count + 1,
        project=project,
        title=title or f'Experiment {project_experiment_count+1}',
    )


def get_experiments(project=None):
    queryset = Experiment.objects.all()
    filters = dict(project=project and project.pk)
    return ExperimentFilter(filters, queryset).qs



from itertools import cycle

import pytest
from model_bakery import baker

from huscy.projects.services import get_data_acquisition_methods

pytestmark = pytest.mark.django_db


@pytest.fixture
def data_acquisition_methods():
    projects = baker.make('projects.Project', _quantity=2)
    experiments = baker.make('projects.Experiment', project=cycle(projects), _quantity=2)
    sessions = baker.make('projects.Session', experiment=cycle(experiments), _quantity=2)
    return baker.make('projects.DataAcquisitionMethod', session=cycle(sessions), _quantity=4)


def test_get_data_acquisition_methods(data_acquisition_methods):
    result = get_data_acquisition_methods()

    assert len(result) == 4
    assert list(result) == data_acquisition_methods


def test_get_data_acquisition_methods_filtered_by_project(data_acquisition_methods):
    project = data_acquisition_methods[0].session.experiment.project

    result = get_data_acquisition_methods(project)

    assert len(result) == 2
    assert list(result) == [data_acquisition_methods[0], data_acquisition_methods[2]]



from huscy.projects.services import update_membership


def test_update_membership_make_user_to_coordinator(membership):
    assert not membership.user.has_perm('change_project', membership.project)

    update_membership(membership, is_coordinator=True, has_write_permission=True)

    assert membership.user.has_perm('change_project', membership.project)
    membership.refresh_from_db()
    assert membership.is_coordinator is True


def test_update_membership_ignores_write_permission_flag_if_is_coordinator_is_true(membership):
    assert not membership.user.has_perm('change_project', membership.project)

    update_membership(membership, is_coordinator=True, has_write_permission=False)

    assert membership.user.has_perm('change_project', membership.project)


def test_update_membership_set_write_permission(membership):
    assert not membership.user.has_perm('change_project', membership.project)

    update_membership(membership, is_coordinator=False, has_write_permission=True)

    assert membership.user.has_perm('change_project', membership.project)
    membership.refresh_from_db()
    assert membership.is_coordinator is False


def test_update_membership_remove_write_permission(membership):
    assert not membership.user.has_perm('change_project', membership.project)

    update_membership(membership, is_coordinator=False, has_write_permission=False)

    assert not membership.user.has_perm('change_project', membership.project)
    membership.refresh_from_db()
    assert membership.is_coordinator is False



from datetime import timedelta

from django.conf import settings
from django.db import models

from .experiment import Experiment


class Session(models.Model):
    experiment = models.ForeignKey(Experiment, on_delete=models.CASCADE, related_name='sessions')
    title = models.CharField(max_length=64)
    order = models.PositiveSmallIntegerField()

    setup_time = models.DurationField(default=timedelta())
    duration = models.DurationField()
    teardown_time = models.DurationField(default=timedelta())

    operator = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.PROTECT)

    max_number_of_participants = models.PositiveIntegerField(default=1)



import pytest

from rest_framework.reverse import reverse
from rest_framework.status import HTTP_200_OK, HTTP_403_FORBIDDEN

from django.contrib.auth.models import Permission

pytestmark = pytest.mark.django_db


def test_admin_user_can_update_research_units(admin_client, admin_user, research_unit):
    response = update_research_unit(admin_client, research_unit, admin_user)

    assert response.status_code == HTTP_200_OK
    assert_research_unit_is_updated(research_unit)


def test_admin_user_can_partial_update_research_units(admin_client, research_unit):
    response = partial_update_research_unit(admin_client, research_unit)

    assert response.status_code == HTTP_200_OK
    assert_research_unit_is_updated(research_unit)


def test_user_with_permission_can_update_research_units(client, user, research_unit):
    update_permission = Permission.objects.get(codename='change_researchunit')
    user.user_permissions.add(update_permission)

    response = update_research_unit(client, research_unit, user)

    assert response.status_code == HTTP_200_OK
    assert_research_unit_is_updated(research_unit)


def test_user_with_permission_can_partial_update_research_units(client, user, research_unit):
    update_permission = Permission.objects.get(codename='change_researchunit')
    user.user_permissions.add(update_permission)

    response = partial_update_research_unit(client, research_unit)

    assert response.status_code == HTTP_200_OK
    assert_research_unit_is_updated(research_unit)


def test_user_without_permission_cannot_update_research_units(client, user, research_unit):
    response = update_research_unit(client, research_unit, user)

    assert response.status_code == HTTP_403_FORBIDDEN
    assert_research_unit_did_not_change(research_unit)


def test_user_without_permission_cannot_partial_update_research_units(client, research_unit):
    response = partial_update_research_unit(client, research_unit)

    assert response.status_code == HTTP_403_FORBIDDEN
    assert_research_unit_did_not_change(research_unit)


def test_anonymous_user_cannot_update_research_units(anonymous_client, user, research_unit):
    response = update_research_unit(anonymous_client, research_unit, user)

    assert response.status_code == HTTP_403_FORBIDDEN
    assert_research_unit_did_not_change(research_unit)


def test_anonymous_user_cannot_partial_update_research_units(anonymous_client, research_unit):
    response = partial_update_research_unit(anonymous_client, research_unit)

    assert response.status_code == HTTP_403_FORBIDDEN
    assert_research_unit_did_not_change(research_unit)


def update_research_unit(client, research_unit, principal_investigator):
    return client.put(
        reverse('researchunit-detail', kwargs=dict(pk=research_unit.id)),
        data=dict(
            name='research_unit_2',
            code='RU2',
            principal_investigator=principal_investigator.pk
        )
    )


def partial_update_research_unit(client, research_unit):
    return client.patch(
        reverse('researchunit-detail', kwargs=dict(pk=research_unit.id)),
        data=dict(name='research_unit_2', code='RU2')
    )


def assert_research_unit_is_updated(research_unit):
    research_unit.refresh_from_db()
    assert 'research_unit_2' == research_unit.name


def assert_research_unit_did_not_change(research_unit):
    research_unit.refresh_from_db()
    assert 'research_unit_1' == research_unit.name



from enum import Enum

from django.db import models

from .session import Session
from .stimulus import Stimulus


class DataAcquisitionMethod(models.Model):
    class TYPE(Enum):
        behavioral = ('behav', 'Behavioral')
        biological = ('bio', 'Biological samples')
        eeg = ('eeg', 'Electroencephalography')
        meg = ('meg', 'Magnetoencephalography')
        microscopy = ('micro', 'Microscopy data')
        mri = ('mri', 'Magnetic resonance imaging')
        nirs = ('nirs', 'Near-Infrared Spectroscopy')
        pause = ('pause', 'Pause')
        pet = ('pet', 'Positron-emission tomography')
        physiological = ('phys', 'Physiological measures')
        questionnaire = ('quest', 'Questionnaire')
        screening = ('screen', 'Screening')

        @classmethod
        def get_value(cls, member):
            return cls[member].value[0]

    session = models.ForeignKey(Session, on_delete=models.CASCADE)
    order = models.PositiveSmallIntegerField()

    stimulus = models.ForeignKey(Stimulus, on_delete=models.SET_NULL, null=True)

    type = models.CharField(max_length=16, choices=[x.value for x in TYPE])

    location = models.CharField(max_length=126)


class Behavioral(models.Model):
    method = models.ForeignKey(DataAcquisitionMethod, on_delete=models.CASCADE,
                               related_name='behaviorals')


class Biological(models.Model):
    method = models.ForeignKey(DataAcquisitionMethod, on_delete=models.CASCADE,
                               related_name='biologicals')


class EEG(models.Model):
    method = models.ForeignKey(DataAcquisitionMethod, on_delete=models.CASCADE, related_name='eeg')


class MEG(models.Model):
    method = models.ForeignKey(DataAcquisitionMethod, on_delete=models.CASCADE, related_name='meg')


class Microscopy(models.Model):
    method = models.ForeignKey(DataAcquisitionMethod, on_delete=models.CASCADE,
                               related_name='microscopys')


class MRI(models.Model):
    method = models.ForeignKey(DataAcquisitionMethod, on_delete=models.CASCADE, related_name='mri')


class NIRS(models.Model):
    method = models.ForeignKey(DataAcquisitionMethod, on_delete=models.CASCADE, related_name='nirs')


class Pause(models.Model):
    method = models.ForeignKey(DataAcquisitionMethod, on_delete=models.CASCADE,
                               related_name='pause')


class PET(models.Model):
    method = models.ForeignKey(DataAcquisitionMethod, on_delete=models.CASCADE, related_name='pet')


class Physiological(models.Model):
    method = models.ForeignKey(DataAcquisitionMethod, on_delete=models.CASCADE,
                               related_name='physiologicals')


class Questionnaire(models.Model):
    method = models.ForeignKey(DataAcquisitionMethod, on_delete=models.CASCADE,
                               related_name='questionnaires')


class Screening(models.Model):
    method = models.ForeignKey(DataAcquisitionMethod, on_delete=models.CASCADE,
                               related_name='screenings')



import pytest

from rest_framework.reverse import reverse
from rest_framework.status import HTTP_200_OK, HTTP_403_FORBIDDEN

from django.contrib.auth.models import Permission

from huscy.projects.models import Project

pytestmark = pytest.mark.django_db


def test_admin_user_can_view_all_projects(admin_client, projects):
    response = list_projects(admin_client)

    assert response.status_code == HTTP_200_OK
    assert_get_all_projects(response)


def test_user_with_permission_can_view_private_projects(client, user, projects):
    can_view_private_projects = Permission.objects.get(codename='can_view_private_projects')
    user.user_permissions.add(can_view_private_projects)

    response = list_projects(client)

    assert response.status_code == HTTP_200_OK
    assert_get_all_projects(response)


def test_user_without_permission_can_only_view_public_projects(client, projects):
    response = list_projects(client)

    assert response.status_code == HTTP_200_OK
    assert_get_only_public_projects(response)


def test_anonymous_user_cannot_view_projects(anonymous_client, projects):
    response = list_projects(anonymous_client)

    assert response.status_code == HTTP_403_FORBIDDEN


def list_projects(client):
    return client.get(reverse('project-list'))


def retrieve_project(client, project):
    return client.get(reverse('project-detail', kwargs=dict(pk=project.pk)))


def assert_get_all_projects(response):
    assert len(response.json()) == 10


def assert_get_only_public_projects(response):
    public = Project.VISIBILITY.get_value('public')

    assert len(response.json()) == 5
    assert all([project['visibility'] == public for project in response.json()])



import pytest

from rest_framework.reverse import reverse
from rest_framework.status import HTTP_200_OK, HTTP_403_FORBIDDEN


def test_admin_user_can_update_experiments(admin_client, experiment):
    response = update_experiment(admin_client, experiment)

    assert_success(response)
    assert_experiment_is_updated(experiment)


def test_user_without_permission_can_update_experiments(client, experiment):
    response = update_experiment(client, experiment)

    assert_success(response)
    assert_experiment_is_updated(experiment)


@pytest.mark.django_db
def test_anonymous_user_cannot_update_experiments(anonymous_client, experiment):
    response = update_experiment(anonymous_client, experiment)

    assert_permission_denied(response)
    assert_experiment_not_updated(experiment)


def update_experiment(client, experiment):
    return client.put(
        reverse('experiment-detail', kwargs=dict(pk=experiment.pk)),
        data=dict(
            description='new description',
            order=experiment.order,
            project=experiment.project.pk,
            title='new title',
        )
    )


def assert_success(response):
    assert response.status_code == HTTP_200_OK


def assert_permission_denied(response):
    assert response.status_code == HTTP_403_FORBIDDEN


def assert_experiment_is_updated(experiment):
    experiment.refresh_from_db()
    assert 'new description' == experiment.description
    assert 'new title' == experiment.title


def assert_experiment_not_updated(experiment):
    experiment.refresh_from_db()
    assert 'description' == experiment.description



import logging

from .memberships import add_member, remove_member
from huscy.projects.models import Project

logger = logging.getLogger('projects')


def create_project(title, research_unit, principal_investigator, creator, visibility,
                   local_id=None, description=''):
    if local_id is None:
        local_id = Project.objects.get_next_local_id(research_unit)

    project = Project.objects.create(
        description=description,
        local_id=local_id,
        principal_investigator=principal_investigator,
        research_unit=research_unit,
        title=title,
        visibility=visibility
    )
    logger.info('Project id:%d, local_id_name:%s, title:%s reserch_unit:%s has been created',
                project.id, project.local_id_name, project.title, project.research_unit.name)

    if principal_investigator != creator:
        add_member(project, creator, is_coordinator=True)

    return project


def delete_project(project):
    map(remove_member, project.membership_set.all())
    logger.info('All members from project <id: %d> removed', project.id)

    project.delete()
    logger.info('Project id:%d, local_id_name:%s, title:%s research_unit:%s has been deleted',
                project.id, project.local_id_name, project.title, project.research_unit.name)


def get_projects(user):
    queryset = Project.objects

    if not user.has_perm('projects.can_view_private_projects'):
        public = Project.VISIBILITY.get_value('public')
        queryset = queryset.filter(visibility=public)

    return queryset.all()



from huscy.projects.models import Session


def create_session(experiment, duration, operator, title='', order=0):
    experiment_session_count = experiment.sessions.count()
    project_session_count = Session.objects.filter(experiment__project=experiment.project).count()

    return Session.objects.create(
        duration=duration,
        experiment=experiment,
        operator=operator,
        order=order or experiment_session_count + 1,
        title=title or f'Session {project_session_count + 1}'
    )


def get_sessions():
    return Session.objects.all()



import pytest
from model_bakery import baker

from rest_framework.reverse import reverse
from rest_framework.status import HTTP_200_OK, HTTP_400_BAD_REQUEST, HTTP_403_FORBIDDEN

from django.contrib.auth.models import Permission

from huscy.projects.models import Project

pytestmark = pytest.mark.django_db


def test_admin_user_can_update_projects(admin_client, public_project):
    response = update_project(admin_client, public_project)

    assert response.status_code == HTTP_200_OK
    assert_project_is_updated(public_project)


def test_admin_user_can_partial_update_projects(admin_client, public_project):
    response = partial_update_project(admin_client, public_project)

    assert response.status_code == HTTP_200_OK
    assert_project_is_updated(public_project)


def test_user_with_permission_can_update_projects(client, user, public_project):
    update_permission = Permission.objects.get(codename='change_project')
    user.user_permissions.add(update_permission)

    response = update_project(client, public_project)

    assert response.status_code == HTTP_200_OK
    assert_project_is_updated(public_project)


def test_user_cannot_update_local_id_to_other_existing_local_id(client, user, project):
    update_permission = Permission.objects.get(codename='change_project')
    user.user_permissions.add(update_permission)
    other = baker.make('projects.Project',
                       research_unit=project.research_unit,
                       local_id=project.local_id + 1)
    response = update_project(client, project, local_id=other.local_id)

    assert response.status_code == HTTP_400_BAD_REQUEST, response.json()
    assert_project_did_not_change(project)


def test_user_can_update_local_id_to_unused_local_id(client, user, project):
    update_permission = Permission.objects.get(codename='change_project')
    user.user_permissions.add(update_permission)

    assert not Project.objects.filter(local_id=project.local_id + 1,
                                      research_unit=project.research_unit).exists()
    response = update_project(client, project, local_id=project.local_id + 1)

    assert response.status_code == HTTP_200_OK

    assert not Project.objects.filter(local_id=project.local_id,
                                      research_unit=project.research_unit).exists()


def test_user_with_permission_can_partial_update_research_units(client, user, public_project):
    update_permission = Permission.objects.get(codename='change_project')
    user.user_permissions.add(update_permission)

    response = partial_update_project(client, public_project)

    assert response.status_code == HTTP_200_OK
    assert_project_is_updated(public_project)


def test_user_without_permission_cannot_update_projects(client, public_project):
    response = update_project(client, public_project)

    assert response.status_code == HTTP_403_FORBIDDEN
    assert_project_did_not_change(public_project)


def test_user_without_permission_cannot_partial_update_projects(client, public_project):
    response = partial_update_project(client, public_project)

    assert response.status_code == HTTP_403_FORBIDDEN
    assert_project_did_not_change(public_project)


def test_anonymous_user_cannot_update_projects(anonymous_client, public_project):
    response = update_project(anonymous_client, public_project)

    assert response.status_code == HTTP_403_FORBIDDEN
    assert_project_did_not_change(public_project)


def test_anonymous_user_cannot_partial_update_projects(anonymous_client, public_project):
    response = partial_update_project(anonymous_client, public_project)

    assert response.status_code == HTTP_403_FORBIDDEN
    assert_project_did_not_change(public_project)


def update_project(client, project, local_id=None):
    return client.put(
        reverse('project-detail', kwargs=dict(pk=project.pk)),
        data=dict(
            description=project.description,
            principal_investigator=project.principal_investigator.id,
            research_unit=project.research_unit.pk,
            title='new_project_title',
            local_id=local_id or project.local_id,
            visibility=project.visibility,
        )
    )


def partial_update_project(client, project):
    return client.patch(
        reverse('project-detail', kwargs=dict(pk=project.pk)),
        data=dict(title='new_project_title')
    )


def assert_project_is_updated(project):
    project.refresh_from_db()
    assert 'new_project_title' == project.title


def assert_project_did_not_change(project):
    project.refresh_from_db()
    assert 'project_title' == project.title



from rest_framework.reverse import reverse
from rest_framework.status import HTTP_405_METHOD_NOT_ALLOWED


def test_retrieve_experiments_is_not_provided(client, experiment):
    response = retrieve_experiment(client, experiment)

    assert_method_not_allowed(response)


def test_list_experiments_is_not_provided(client):
    response = list_experiments(client)

    assert_method_not_allowed(response)


def retrieve_experiment(client, experiment):
    return client.get(reverse('experiment-detail', kwargs=dict(pk=experiment.pk)))


def list_experiments(client):
    return client.get(reverse('experiment-list'))


def assert_method_not_allowed(response):
    assert response.status_code == HTTP_405_METHOD_NOT_ALLOWED



from rest_framework import serializers

from huscy.projects.models import DataAcquisitionMethod
from huscy.projects.services import add_data_acquisition_method


class DataAcquisitionMethodSerializer(serializers.ModelSerializer):
    class Meta:
        model = DataAcquisitionMethod
        fields = (
            'id',
            'location',
            'order',
            'session',
            'type',
        )

    def create(self, validated_data):
        return add_data_acquisition_method(**validated_data)



from rest_framework.reverse import reverse
from rest_framework.status import HTTP_204_NO_CONTENT, HTTP_403_FORBIDDEN


def test_admin_user_can_delete_memberships(admin_client, membership):
    response = delete_membership(admin_client, membership)

    assert response.status_code == HTTP_204_NO_CONTENT


def test_user_without_permission_can_delete_memberships(client, membership):
    response = delete_membership(client, membership)

    assert response.status_code == HTTP_204_NO_CONTENT


def test_anonymous_user_cannot_delete_memberships(anonymous_client, membership):
    response = delete_membership(anonymous_client, membership)

    assert response.status_code == HTTP_403_FORBIDDEN


def delete_membership(client, membership):
    return client.delete(reverse('membership-detail', kwargs=dict(pk=membership.id)))



import pytest

from huscy.projects.services import add_data_acquisition_method

pytestmark = pytest.mark.django_db


def test_with_order_attribute(data_acquisition_method):
    session = data_acquisition_method.session
    method = add_data_acquisition_method(session, 'eeg', 'location', order=5)

    assert method.session == session
    assert method.type == 'eeg'
    assert method.location == 'location'
    assert method.order == 5


def test_without_order_attribute(data_acquisition_method):
    session = data_acquisition_method.session
    method = add_data_acquisition_method(session, 'eeg', 'location')

    assert method.session == session
    assert method.type == 'eeg'
    assert method.location == 'location'
    assert method.order == 2



from datetime import timedelta

import pytest
from model_bakery import baker

from huscy.projects.services import create_session


@pytest.fixture
def duration():
    return timedelta(hours=1)


def test_create_session(user, experiment, duration):
    session = create_session(experiment, duration, user)

    assert session.experiment == experiment
    assert session.duration == duration
    assert session.operator == user
    assert session.order == 1
    assert session.title == 'Session 1'


def test_create_session_to_existing_ones(user, experiment, duration):
    baker.make('projects.Session', experiment=experiment, _quantity=2)

    session = create_session(experiment, duration, user)

    assert session.order == 3
    assert session.title == 'Session 3'


def test_create_session_to_second_experiment(user, experiment, duration):
    baker.make('projects.Session', experiment=experiment, _quantity=2)
    second_experiment = baker.make('projects.Experiment', project=experiment.project)

    session = create_session(second_experiment, duration, user)

    assert session.order == 1
    assert session.title == 'Session 3'


def test_create_session_with_title(user, experiment, duration):
    session = create_session(experiment, duration, user, title='new session')

    assert session.title == 'new session'


def test_create_session_with_order(user, experiment, duration):
    session = create_session(experiment, duration, user, order=5)

    assert session.order == 5
